@@ -20,7 +20,7 @@ MyNamedTuple = namedtuple('MyNamedTuple', 'x y')
 def test_types():
     assert_equal(packb(dict()), packb(MyDict()))
     assert_equal(packb(list()), packb(MyList()))
-    assert_equal(packb(MyNamedTuple(1,2)), (1,2))
+    assert_equal(packb(MyNamedTuple(1,2)), packb((1,2)))
 
 
 if __name__ == '__main__':
